import os
import json

from openpyxl import load_workbook
from excel_to_json.sheet import Sheet
import excel_to_json.columntitle as columntitle
import excel_to_json.sheetname as sheetname
from excel_to_json.date_encoder import DateEncoder


class Book:
    def __init__(self):
        self.sheets = []

    def parse_file(self, file_path):
        wb = load_workbook(file_path)
        for ws in wb:
            sheet = Sheet(ws.title)
            sheet.parse(ws)
            self.sheets.append(sheet)
        # deal the joint
        for sheet in self.sheets:
            for target, join in sheet.joins.items():
                target_sheet_name = columntitle.join_target_sheet(target)
                target_column = columntitle.join_target_column(target)
                target_sheet = self.get_sheet(target_sheet_name)
                for join_identify, datas in join.items():
                    setter = target_sheet.get_anchor_setter(target_column, join_identify)
                    if setter:
                        for data in datas:
                            setter(data)

    def out_file(self, path):
        # check predefined out file suffix, like .json or .yaml
        data = None
        for sheet in filter(lambda s: sheetname.is_outfile(s.title), self.sheets):
            try:
                data = self.get_data(sheet)
                f = open(os.path.join(path, sheetname.out_file_name(sheet.title)), 'w+', encoding='utf-8')
                f.write(data)
                f.close()
            except Exception as err:
                print(err)

    def out_console(self):
        # check predefined out file suffix, like .json or .yaml
        data = None
        for sheet in self.sheets:
            try:
                data = self.get_data(sheet)
                print(data)
            except Exception as err:
                print(err)

    def get_sheet(self, sheet_name):
        return next(filter(lambda s: sheetname.shortname(s.title) == sheet_name, self.sheets), None)

    @staticmethod
    def get_data(sheet):
        data = None
        if sheetname.is_json(sheet.title):
            data = Book.get_json_data(sheet)
        # can add yaml support
        return data

    @staticmethod
    def get_json_data(sheet):
        if not sheetname.is_array(sheet.title) and len(sheet.datas) == 1:
            data = json.dumps(sheet.datas[0], cls=DateEncoder, ensure_ascii=False)
        else:
            data = json.dumps(sheet.datas, cls=DateEncoder, ensure_ascii=False)
        return data




import excel_to_json.columntitle as columntitle


class Sheet:
    def __init__(self, title):
        self.title = title
        self.datas = []
        self.anchors = {}
        self.joins = {}

    def parse(self, sheet):
        for column in filter(lambda col: len(col) > 1 and columntitle.is_valid(col[0].value), sheet.columns):
            title = column[0].value.strip()
            decorator = columntitle.parse_decorator(title)
            for row_index in range(1, len(column)):
                if decorator is columntitle.Decorator.Empty:
                    self.parse_data(row_index - 1, title, column[row_index].value)
                elif decorator is columntitle.Decorator.Anchor:
                    self.parse_anchor(row_index - 1, title, column[row_index].value)
                elif decorator is columntitle.Decorator.Join:
                    self.parse_join(row_index - 1, title, column[row_index].value)

    def lazy_get_data(self, row_index):
        while len(self.datas) <= row_index:
            self.datas.append({})
        return self.datas[row_index]

    def parse_data(self, row_index, title, val):
        if val is not None and str(val).strip() != '':
            component_container = self.lazy_get_data(row_index)
            components = columntitle.parse_components(title)
            for type, name in components:
                if type is columntitle.Type.Object:
                    if name not in component_container:
                        component_container[name] = {}
                    component_container = component_container[name]
                elif type is columntitle.Type.Array:
                    component_container[name] = val.split(',')
                elif type is columntitle.Type.Attri:
                    component_container[name] = val

    def parse_join(self, row_index, title, val):
        if val is not None and str(val).strip() != '':
            pure_title = columntitle.title_without_decorator(title)
            row_data = self.lazy_get_data(row_index)
            if pure_title not in self.joins:
                self.joins[pure_title] = {}
            if val not in self.joins[pure_title]:
                self.joins[pure_title][val] = []
            self.joins[pure_title][val].append(row_data)

    def parse_anchor(self, row_index, title, val):
        if val is not None and str(val).strip() != '':
            # strip decorator
            pure_title = columntitle.title_without_decorator(title)
            anchor_setter = None
            component_container = self.lazy_get_data(row_index)
            components = columntitle.parse_components(pure_title)
            for type, name in components:
                if type is columntitle.Type.Object:
                    component_container[name] = {}
                    component_container = component_container[name]
                elif type is columntitle.Type.Array:
                    component_container[name] = []

                    def setter(data):
                        component_container[name].append(data)
                    anchor_setter = setter
                elif type is columntitle.Type.Attri:
                    def setter(data):
                        component_container[name] = data
                    anchor_setter = setter
            if pure_title not in self.anchors:
                self.anchors[pure_title] = {}
            self.anchors[pure_title][val] = anchor_setter

    def get_anchor_setter(self, title, val):
        # strip decorator
        pure_title = columntitle.title_without_decorator(title)
        setter = None
        if pure_title in self.anchors:
            title_dict = self.anchors[pure_title]
            if val in title_dict:
                setter = title_dict[val]
        return setter

    def get_join(self, join_target, identify):
        return self.joins[columntitle.title_without_decorator(join_target)][identify]



import os.path
import sys
import json

from excel_to_json.book import Book


def convert_xlsl(excel_file):
    file_dir = os.path.dirname(excel_file)
    workbook = Book()
    workbook.parse_file(excel_file)
    workbook.out_file(file_dir)


def convert_folder(folder_path):
    paths = os.listdir(folder_path)
    for path in paths:
        if os.path.isdir(path):
            convert_folder(path)
        elif os.path.isfile(path) and path.endswith('.xlsx'):
            convert_xlsl(path)


def excel_to_json():
    input_paths = sys.argv[1:]
    for path in input_paths:
        if os.path.isdir(path):
            convert_folder(path)
        elif os.path.isfile(path) and path.endswith('.xlsx'):
            convert_xlsl(path)



def shortname(sheetname):
    index = sheetname.find('.')
    return sheetname if index == -1 else sheetname[:index]

def is_outfile(sheetname):
    return is_json(sheetname)

def is_json(sheetname):
    return sheetname.find('.json') != -1


def is_array(sheetname):
    return sheetname.find('@') != -1

def out_file_name(sheetname):
    return sheetname.rstrip('@')


from enum import Enum
from collections import namedtuple


class Decorator(Enum):
    Empty = 1
    Anchor = 2
    Join = 3


def parse_decorator(title):
    decorator = Decorator.Empty
    if title.find('#anchor') != -1:
        decorator = Decorator.Anchor
    elif title.find('#join') != -1:
        decorator = Decorator.Join
    else:
        decorator = Decorator.Empty
    return decorator


def title_without_decorator(title):
    decorator_index = title.find('#')
    return title[:] if decorator_index == -1 else title[:decorator_index]


def join_target_sheet(title):
    sheet_separator_index = title.find('/')
    if sheet_separator_index != -1:
        return title[:sheet_separator_index]
    return None


def join_target_column(title):
    begin_index = 0
    sheet_separator_index = title.find('/')
    if sheet_separator_index != -1:
        begin_index = sheet_separator_index + 1
    end_index = title.find('#')
    return title_without_decorator(title[begin_index:] if end_index == -1 else title[begin_index:end_index])


class Type(Enum):
    Attri = 1,
    Array = 2,
    Object = 3,


Component = namedtuple('Component', ['type', 'name'])


def parse_components(title):
    components = []
    decorator_index = title.find('#')
    title_without_decorator = title if decorator_index == -1 else title[:decorator_index]
    raw_components = title_without_decorator.split('.')
    if len(raw_components) > 0:
        for raw_component in raw_components[:-1]:
            components.append(Component(Type.Object, raw_component))
        last_raw_component = raw_components[-1]
        if last_raw_component.endswith('@'):
            components.append(Component(Type.Array, last_raw_component[:-1]))
        else:
            components.append(Component(Type.Attri, last_raw_component))
    return components


def is_valid(title):
    return title is not None and title.strip() != "" and not title.startswith('#')



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='excel_to_json',
      version='1.4',
      description='excel to json',
      long_description=readme(),
      classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Text Processing',
      ],
      keywords='excel to json',
      url='http://github.com/wang37921/excel_to_json',
      author='Wang Cong',
      author_email='wang37921@163.com',
      license='MIT',
      packages=['excel_to_json'],
      install_requires=[
          'openpyxl',
      ],
      test_suite='nose.collector',
      tests_require=['nose', 'nose-cover3'],
      entry_points={
          'console_scripts': ['excel-to-json=excel_to_json.command_line:excel_to_json'],
      },
      include_package_data=True,
      zip_safe=False)